#!/usr/bin/env python
import resource
import time

import angr
from tracer import QEMURunner
import angr.engines.vex.statements.base as vex_statements
import angr.engines.vex.engine as vex_engine

LOG_FILE = "/tmp/angr_instruction_log.txt"
LOG_INTERVAL_SEC = 60
TIMEOUT_SEC = 10 * 60

def run(binary, input_data):
    """
    Run the given binary with the provided data on stdin and dump inputs that
    divert execution at one conditional each.
    """
    runner = QEMURunner(binary, input_data)
    proj = angr.Project(binary)

    with open(LOG_FILE, 'w') as f:
        f.write("0 0 0\n")

    # print "Running concretely with input %s" % input_data.strip()
    if proj.loader.main_object.os == 'cgc':
        proj.simos.syscall_library.update(angr.SIM_LIBRARIES['cgcabi_tracer'])
        init_state = proj.factory.entry_state(stdin=angr.SimFileStream,
                                              flag_page=runner.magic,
                                              add_options=angr.options.unicorn)
    else:
        init_state = proj.factory.full_init_state(stdin=angr.SimFileStream)

    init_state.preconstrainer.preconstrain_file(input_data,
                                                init_state.posix.stdin, True)

    print "Following with the symbolic executor..."
    simgr = proj.factory.simulation_manager(init_state, save_unsat=True,
                                            hierarchy=False,
                                            save_unconstrained=runner.crash_mode)

    if runner.crash_mode:
        simgr.use_technique(angr.exploration_techniques.CrashMonitor(
            trace=runner.trace, crash_addr=runner.crash_addr))

    simgr.use_technique(angr.exploration_techniques.Tracer(trace=runner.trace))
    simgr.use_technique(angr.exploration_techniques.Oppologist())
    simgr.use_technique(angr.exploration_techniques.DrillerCore(trace=runner.trace))

    start_time = last_log = time.time()
    while simgr.active and simgr.one_active.globals["bb_cnt"] < len(runner.trace) - 1:
        simgr.step()

        now = time.time()
        if now - last_log > LOG_INTERVAL_SEC:
            with open(LOG_FILE, 'a') as f:
                f.write("%d %d %d\n" %
                        (vex_statements.executed_statements,
                         now - start_time,
                         resource.getrusage(resource.RUSAGE_SELF).ru_maxrss))
            last_log = now

        if now - start_time > TIMEOUT_SEC + 1:
            print "Time is up!"
            return

        if "diverted" in simgr.stashes:
            while simgr.diverted:
                print "Discarding diverted state"
                simgr.diverted.pop()

    # We may use this to get the diverting inputs:
    # generated = state.posix.stdin.load(0, state.posix.stdin.pos)
    # generated = state.solver.eval(generated, cast_to=bytes)

    with open(LOG_FILE, 'a') as f:
        f.write("%d %d %d\n" %
                (vex_statements.executed_statements,
                 vex_engine.symbolic_time,
                 resource.getrusage(resource.RUSAGE_SELF).ru_maxrss))

if __name__ == "__main__":
    import sys
    if len(sys.argv) != 3:
        print "Usage: %s <binary> <input>"
        sys.exit(-1)

    with open(sys.argv[2]) as f:
        data = f.read()

    mem_limit_gb = 25
    print "Setting memory limit to %d GB" % mem_limit_gb
    mem_limit_bytes = mem_limit_gb * 1024 * 1024 * 1024
    resource.setrlimit(resource.RLIMIT_AS, (mem_limit_bytes, mem_limit_bytes))

    run(sys.argv[1], data)
